/*
 * Ben NanoNote LCD initialisation, based on uboot-xburst and xburst-tools.
 *
 * Copyright (C) 2001-2002 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "board-display.h"

#include "xburst_types.h"
#include "jzlcd.h"
#include "sdram.h"
#include "board.h"

extern vidinfo_t panel_info;
static unsigned long lcd_base;

static unsigned int get_line_length()
{
	return ALIGN((panel_info.vl_col * NBITS(panel_info.vl_bpix)) / 8, sizeof(u32));
}

static u32 *get_pixel32(unsigned short h, unsigned short v)
{
	return (u32 *) (lcd_base + v * get_line_length()) + h;
}

static u16 *get_pixel16(unsigned short h, unsigned short v)
{
	return (u16 *) (lcd_base + v * get_line_length()) + h;
}

static u8 *get_pixel8(unsigned short h, unsigned short v)
{
	return (u8 *) (lcd_base + v * get_line_length()) + h;
}

static u8 *get_pixel4(unsigned short h, unsigned short v)
{
	return (u8 *) (lcd_base + v * get_line_length()) + h / 2;
}

static inline unsigned short div(unsigned short num, unsigned short denom, unsigned short scale)
{
	return (scale * num) / denom;
}

static unsigned long pixel(u8 r, u8 g, u8 b, u8 rmax, u8 gmax, u8 bmax, u8 rshift, u8 gshift, u8 bshift)
{
	return (div(r, 255, rmax) << rshift) | (div(g, 255, gmax) << gshift) | (div(b, 255, bmax) << bshift);
}

static void get_colour(unsigned short h, unsigned short v, u8 rgb[], unsigned short pixel_type)
{
	unsigned short v_max = panel_info.vl_row;
	unsigned short h_max = panel_info.vl_col;

	rgb[(pixel_type - 1) % 3] = div(h, h_max, 255);
	rgb[pixel_type % 3] = div(v, v_max, 255);
	rgb[(pixel_type + 1) % 3] = (rgb[0] + rgb[1]) / 2;
}

static void test_pixel32(unsigned short h, unsigned short v, unsigned short pixel_type)
{
	u32 *pix = get_pixel32(h, v);
	u8 rgb[3];

	get_colour(h, v, rgb, pixel_type);
	*pix = (u32) pixel(rgb[0], rgb[1], rgb[2], 255, 255, 255, 16, 8, 0);
}

static void test_pixel16_565(unsigned short h, unsigned short v, unsigned short pixel_type)
{
	u16 *pix = get_pixel16(h, v);
	u8 rgb[3];

	get_colour(h, v, rgb, pixel_type);
	*pix = (u16) pixel(rgb[0], rgb[1], rgb[2], 31, 63, 31, 11, 5, 0);
}

static void test_pixel8(unsigned short h, unsigned short v, unsigned short pixel_type)
{
	u8 *pix = get_pixel8(h, v);
	u8 rgb[3];

	get_colour(h, v, rgb, pixel_type);
	*pix = (u8) pixel(rgb[0], rgb[1], rgb[2], 7, 7, 3, 5, 2, 0);
}

static void test_pixel4(unsigned short h, unsigned short v, unsigned short pixel_type)
{
	u8 *pix = get_pixel4(h, v);
	u8 mask = h & 1 ? 0xf0 : 0x0f;
	u8 rgb[3];

	get_colour(h, v, rgb, pixel_type);
	*pix = (*pix & mask) | ((u8) pixel(rgb[0], rgb[1], rgb[2], 1, 2, 1, 3, 1, 0) << (h & 1 ? 0 : 4));
}

void test_pixel(unsigned short h, unsigned short v, unsigned short pixel_type)
{
	switch (panel_info.vl_bpix)
	{
		case LCD_COLOR32:
		test_pixel32(h, v, pixel_type);
		break;

		case LCD_COLOR8:
		test_pixel8(h, v, pixel_type);
		break;

		case LCD_COLOR4:
		test_pixel4(h, v, pixel_type);
		break;

		case LCD_COLOR16:
		default:
		test_pixel16_565(h, v, pixel_type);
		break;
	}
}

void clear_pixel32(unsigned short h, unsigned short v)
{
	u32 *pix = get_pixel32(h, v);
	*pix = 0;
}

void clear_pixel16(unsigned short h, unsigned short v)
{
	u16 *pix = get_pixel16(h, v);
	*pix = 0;
}

void clear_pixel8(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel8(h, v);
	*pix = 0;
}

void clear_pixel4(unsigned short h, unsigned short v)
{
	u8 *pix = get_pixel4(h, v);
	u8 mask = h & 1 ? 0xf0 : 0x0f;
	*pix = *pix & mask;
}

void clear_pixel(unsigned short h, unsigned short v)
{
	switch (panel_info.vl_bpix)
	{
		case LCD_COLOR32:
		clear_pixel32(h, v);
		break;

		case LCD_COLOR8:
		clear_pixel8(h, v);
		break;

		case LCD_COLOR4:
		clear_pixel4(h, v);
		break;

		case LCD_COLOR16:
		default:
		clear_pixel16(h, v);
		break;
	}
}

void test_pattern()
{
	unsigned short v_max  = panel_info.vl_row;
	unsigned short h_max  = panel_info.vl_col;
	unsigned short v, h;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			test_pixel(h, v, 1);
		}
	}
}

void lcd_clear(unsigned long lcd_base)
{
	unsigned short v_max  = panel_info.vl_row;
	unsigned short h_max  = panel_info.vl_col;
	unsigned short v, h;
	unsigned long *pix = (unsigned long *)lcd_base;

	for (v = 0; v < v_max; v += 1) {
		for (h = 0; h < h_max; h += 1) {
			*pix++ = 0;
		}
	}
}

/* LCD initialisation. */

void lcd_init(void)
{
	__lcd_display_pin_init();
	__lcd_display_on();

	lcd_base = lcd_ctrl_init();
	lcd_clear(lcd_base);
	lcd_enable();
}
